# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
import handler_base
import v2_context as a10


class MemberHandler(handler_base.HandlerBase):

    def _get_ip(self, context, member, use_float=False):
        return self.openstack_manager._get_ip(context, member, use_float)

    def _get_name(self, member, ip_address):
        tenant_label = member.tenant_id[:5]
        addr_label = str(ip_address).replace(".", "_", 4)
        server_name = "_%s_%s_neutron" % (tenant_label, addr_label)
        return server_name

    def _count(self, context, member):
        return self.openstack_manager._count(context, member)

    def create(self, context, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            server_ip = self._get_ip(context, member,
                                     c.device_cfg['use_float'])
            server_name = self._get_name(member, server_ip)

            status = c.client.slb.UP
            if not member.admin_state_up:
                status = c.client.slb.DOWN

            try:
                c.client.slb.server.create(server_name, server_ip)
            except acos_errors.Exists:
                pass

            c.client.slb.service_group.member.create(member.pool.id,
                                                     server_name,
                                                     member.protocol_port,
                                                     status=status)

    def update(self, context, old_member, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            server_ip = self._get_ip(context, member,
                                     c.device_cfg['use_float'])
            server_name = self._get_name(member, server_ip)

            status = c.client.slb.UP
            if not member.admin_state_up:
                status = c.client.slb.DOWN

            c.client.slb.service_group.member.update(member.pool.id,
                                                     server_name,
                                                     member.protocol_port,
                                                     status)

    def _delete(self, c, context, member):
        server_ip = self._get_ip(context, member, c.device_cfg['use_float'])
        server_name = self._get_name(member, server_ip)

        if self._count(context, member) > 1:
            c.client.slb.service_group.member.delete(member.pool.id,
                                                     server_name,
                                                     member.protocol_port)
        else:
            c.client.slb.server.delete(server_name)

    def delete(self, context, member):
        with a10.A10DeleteContext(self, context, member) as c:
            self._delete(c, context, member)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.v2.v2_context as a10

import test_base


class FakeException(Exception):
    pass


class TestA10Context(test_base.UnitTestBase):

    def setUp(self):
        super(TestA10Context, self).setUp()
        self.handler = self.a.pool
        self.ctx = None
        self.m = test_base.FakeModel()

    def test_context(self):
        with a10.A10Context(self.handler, self.ctx, self.m) as c:
            self.empty_mocks()
            c
        self.empty_close_mocks()

    def test_context_e(self):
        try:
            with a10.A10Context(self.handler, self.ctx, self.m) as c:
                self.empty_mocks()
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()

    def test_write(self):
        with a10.A10WriteContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.last_client.system.action.write_memory.assert_called_with()
        self.a.last_client.session.close.assert_called_with()

    def test_write_e(self):
        try:
            with a10.A10WriteContext(self.handler, self.ctx, self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()
            pass

    def test_write_status(self):
        with a10.A10WriteStatusContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.openstack_driver.pool.active.assert_called_with(
            None, 'fake-id-001')

    def test_write_status_e(self):
        try:
            with a10.A10WriteStatusContext(self.handler, self.ctx,
                                           self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.a.openstack_driver.pool.failed.assert_called_with(
                None, 'fake-id-001')
            pass

    def test_delete(self):
        with a10.A10DeleteContext(self.handler, self.ctx, self.m) as c:
            c
        self.a.openstack_driver.pool.db_delete.assert_called_with(
            None, 'fake-id-001')

    def test_delete_e(self):
        try:
            with a10.A10DeleteContext(self.handler, self.ctx, self.m) as c:
                c
                raise FakeException()
        except FakeException:
            self.empty_close_mocks()
            pass


# Re-run all the context manager tests with appliance partitioning.
class TestA10ContextADP(TestA10Context):

    def setUp(self):
        super(TestA10ContextADP, self).setUp()
        self.reset_v_method('adp')

    def tearDown(self):
        self.reset_v_method('lsi')

    def reset_v_method(self, val):
        for k, v in self.a.config.devices.items():
            v['v_method'] = val

    def empty_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(1, len(self.a.last_client.mock_calls))
        self.a.last_client.system.partition.active.assert_called_with(
            self.m.tenant_id[0:13])

    def empty_close_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(2, len(self.a.last_client.mock_calls))
        self.a.last_client.session.close.assert_called_with()



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import test_base


class TestHM(test_base.UnitTestBase):

    def assert_hm(self, mon_type, method, url, expect_code):
        self.a.openstack_driver.health_monitor.active.assert_called_with(
            None, 'fake-hm-id-001')
        self.a.last_client.slb.hm.create.assert_called_with(
            'fake-hm-id-001', mon_type, 6, 7, 8,
            method=method, url=url, expect_code=expect_code)

    def test_create_ping(self):
        self.a.hm.create(None, test_base.FakeHM('PING'))
        self.assert_hm(self.a.last_client.slb.hm.ICMP, None, None, None)

    def test_create_tcp(self):
        self.a.hm.create(None, test_base.FakeHM('TCP'))
        self.assert_hm(self.a.last_client.slb.hm.TCP, None, None, None)

    def test_create_http(self):
        self.a.hm.create(None, test_base.FakeHM('HTTP'))
        self.assert_hm(self.a.last_client.slb.hm.HTTP, 'GET', '/', '200')

    def test_create_https(self):
        self.a.hm.create(None, test_base.FakeHM('HTTPS'))
        self.assert_hm(self.a.last_client.slb.hm.HTTPS, 'GET', '/', '200')

    def test_create_http_with_pool(self):
        m = test_base.FakeHM('HTTP', pool=mock.MagicMock())
        self.a.hm.create(None, m)
        self.assert_hm(self.a.last_client.slb.hm.HTTP, 'GET', '/', '200')
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='fake-hm-id-001')

    def test_update_tcp(self, m_old=None, m=None):
        if m_old is None:
            m_old = test_base.FakeHM('TCP')
        if m is None:
            m = test_base.FakeHM('TCP')
        m.delay = 20
        self.a.hm.update(None, m_old, m)
        self.a.openstack_driver.health_monitor.active.assert_called_with(
            None, 'fake-hm-id-001')
        self.a.last_client.slb.hm.update.assert_called_with(
            'fake-hm-id-001', self.a.last_client.slb.hm.TCP, 20, 7, 8,
            method=None, url=None, expect_code=None)

    def test_update_tcp_add_pool(self):
        m = test_base.FakeHM('TCP', pool=mock.MagicMock())
        self.test_update_tcp(m=m)
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='fake-hm-id-001')

    def test_update_tcp_delete_pool(self):
        m_old = test_base.FakeHM('TCP', pool=mock.MagicMock())
        self.test_update_tcp(m_old=m_old)
        self.a.last_client.slb.service_group.update.assert_called_with(
            m_old.pool.id, health_monitor='')

    def test_delete(self):
        self.a.hm.delete(None, test_base.FakeHM('HTTP'))
        self.a.openstack_driver.health_monitor.db_delete.assert_called_with(
            None, 'fake-hm-id-001')
        self.a.last_client.slb.hm.delete.assert_called_with('fake-hm-id-001')

    def test_delete_with_pool(self):
        m = test_base.FakeHM('TCP', pool=mock.MagicMock())
        self.a.hm.delete(None, m)
        self.a.openstack_driver.health_monitor.db_delete.assert_called_with(
            None, 'fake-hm-id-001')
        self.a.last_client.slb.service_group.update.assert_called_with(
            m.pool.id, health_monitor='')
        self.a.last_client.slb.hm.delete.assert_called_with('fake-hm-id-001')



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import test_base

import a10_neutron_lbaas.a10_exceptions as a10_ex


class TestVIP(test_base.UnitTestBase):

    def fake_vip(self, pers=None):
        h = {
            'tenant_id': 'ten1',
            'id': 'id1',
            'protocol': 'HTTP',
            'admin_state_up': True,
            'address': '1.1.1.1',
            'protocol_port': '80',
            'pool_id': 'pool1',
        }
        if pers:
            h['session_persistence'] = {'type': pers}
        return h.copy()

    def test_create(self):
        self.a.vip.create(None, self.fake_vip())
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('virtual_server.create' in s)
        self.assertTrue('1.1.1.1' in s)
        self.assertTrue('vport.create' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.assertTrue('get_pool' in s)
        self.assertTrue('HTTP' in s)

    def test_create_pers(self):
        self.a.vip.create(None, self.fake_vip('HTTP_COOKIE'))
        s = str(self.a.last_client.mock_calls)
        self.assertTrue("c_pers_name='id1'" in s)

    def test_create_unsupported(self):
        try:
            self.a.vip.create(None, self.fake_vip('APP_COOKIE'))
        except a10_ex.UnsupportedFeature:
            pass

    def test_update(self):
        self.a.vip.update(None, self.fake_vip(), self.fake_vip())
        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue('vport.update' in s)
        self.assertTrue('id1' in s)
        self.assertTrue('UP' in s)
        self.assertTrue('get_pool' in s)
        self.assertTrue('HTTP' in s)

    def test_delete(self):
        self.a.vip.delete(None, self.fake_vip())
        self.a.last_client.slb.virtual_server.delete.assert_called_with('id1')

    def test_delete_pers(self):
        self.a.vip.delete(None, self.fake_vip('SOURCE_IP'))
        self.a.last_client.slb.virtual_server.delete.assert_called_with('id1')
        z = self.a.last_client.slb.template.src_ip_persistence.delete
        z.assert_called_with('id1')



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import unittest

import mock

import a10_neutron_lbaas.a10_openstack_lb as a10_os


class FakeA10OpenstackLBV1(a10_os.A10OpenstackLBV1):

    def __init__(self, openstack_driver):
        super(FakeA10OpenstackLBV1, self).__init__(mock.MagicMock())

    def _get_a10_client(self, device_info):
        self.device_info = device_info
        self.last_client = mock.MagicMock()
        return self.last_client

    def reset_mocks(self):
        self.openstack_driver = mock.MagicMock()
        self.last_client = self._get_a10_client(self.device_info)
        return self.last_client


class FakeA10OpenstackLBV2(a10_os.A10OpenstackLBV2):

    def __init__(self, openstack_driver):
        super(FakeA10OpenstackLBV2, self).__init__(mock.MagicMock())

    def _get_a10_client(self, device_info):
        self.device_info = device_info
        self.last_client = mock.MagicMock()
        return self.last_client

    def reset_mocks(self):
        self.openstack_driver = mock.MagicMock()
        self.last_client = self._get_a10_client(self.device_info)
        return self.last_client


class UnitTestBase(unittest.TestCase):

    def setUp(self):
        unit_dir = os.path.dirname(__file__)
        unit_config = os.path.join(unit_dir, "unit_config")
        os.environ['A10_CONFIG_DIR'] = unit_config
        if not hasattr(self, 'version') or self.version == 'v2':
            self.a = FakeA10OpenstackLBV2(None)
        else:
            self.a = FakeA10OpenstackLBV1(None)

    def print_mocks(self):
        print("OPENSTACK ", self.a.openstack_driver.mock_calls)
        print("CLIENT ", self.a.last_client.mock_calls)

    def empty_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(0, len(self.a.last_client.mock_calls))

    def empty_close_mocks(self):
        self.print_mocks()
        self.assertEqual(0, len(self.a.openstack_driver.mock_calls))
        self.assertEqual(1, len(self.a.last_client.mock_calls))
        self.a.last_client.session.close.assert_called_with()



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.a10_exceptions as a10_ex

import acos_client.errors as acos_errors
import handler_base
import v2_context as a10


class PoolHandler(handler_base.HandlerBase):

    def _set(self, c, set_method, context, pool):
        lb_methods = {
            'ROUND_ROBIN': c.client.slb.service_group.ROUND_ROBIN,
            'LEAST_CONNECTIONS': c.client.slb.service_group.LEAST_CONNECTION,
            'SOURCE_IP': c.client.slb.service_group.WEIGHTED_LEAST_CONNECTION
        }
        protocols = {
            'HTTP': c.client.slb.service_group.TCP,
            'HTTPS': c.client.slb.service_group.TCP,
            'TCP': c.client.slb.service_group.TCP,
            'UDP': c.client.slb.service_group.UDP
        }

        set_method(pool.id,
                   protocol=protocols[pool.protocol],
                   lb_method=lb_methods[pool.lb_algorithm])

        if pool.sessionpersistence:
            PersistenceHandler(self, c, context, pool).create()

    def create(self, context, pool):
        with a10.A10WriteStatusContext(self, context, pool) as c:
            self._set(c, c.client.slb.service_group.create, context, pool)

    def update(self, context, old_pool, pool):
        with a10.A10WriteStatusContext(self, context, pool) as c:
            self._set(c, c.client.slb.service_group.update, context, pool)

    def delete(self, context, pool):
        with a10.A10DeleteContext(self, context, pool) as c:
            for member in pool.members:
                self.a10_driver.member._delete(c, context, member)

            if pool.healthmonitor:
                self.a10_driver.hm._delete(c, context, pool.healthmonitor)

            c.client.slb.service_group.delete(pool.id)

            if pool.sessionpersistence:
                PersistenceHandler(self, c, context, pool).delete()


class PersistenceHandler(object):

    def __init__(self, pool_handler, c, context, pool):
        self.pool_handler = pool_handler
        self.c = c
        self.context = context
        self.pool = pool
        self.sp = pool.sessionpersistence
        self.name = pool.id

    def create(self):
        methods = {
            'HTTP_COOKIE':
                self.c.client.slb.template.cookie_persistence.create,
            'SOURCE_IP':
                self.c.client.slb.template.src_ip_persistence.create,
        }
        if self.sp.type in methods:
            try:
                methods[self.sp.type](self.name)
            except acos_errors.Exists:
                pass
        else:
            raise a10_ex.UnsupportedFeature()

        if self.pool.listener:
            self.pool_handler.a10_driver.listener._update(self.c, self.context,
                                                          self.pool.listener)

    def delete(self):
        methods = {
            'HTTP_COOKIE':
                self.c.client.slb.template.cookie_persistence.delete,
            'SOURCE_IP':
                self.c.client.slb.template.src_ip_persistence.delete,
        }
        if self.sp.type in methods:
            try:
                methods[self.sp.type](self.name)
            except Exception:
                pass

        if self.pool.listener:
            self.pool_handler.a10_driver.listener._update(self.c, self.context,
                                                          self.pool.listener)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tests.unit.test_base as test_base


class FakeModel(object):

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', 'fake-id-001')
        self.tenant_id = kwargs.get('tenant_id', 'get-off-my-lawn')


class FakeLoadBalancer(FakeModel):

    def __init__(self, listeners=[]):
        super(FakeLoadBalancer, self).__init__()
        self.id = 'fake-lb-id-001'
        self.listeners = listeners
        self.address = '5.5.5.5'
        self.admin_state_up = True


class FakeListener(FakeModel):

    def __init__(self, protocol, port, admin_state_up=True, pool=None,
                 loadbalancer=None):
        super(FakeListener, self).__init__()
        self.id = 'fake-listen-id-001'
        self.protocol = protocol
        self.port = port
        self.admin_state_up = admin_state_up
        self.default_pool = pool
        self.loadbalancer = loadbalancer


class FakePersistence(FakeModel):

    def __init__(self, persistence_type):
        super(FakePersistence, self).__init__()
        self.id = 'fake-pers-id-001'
        self.type = persistence_type


class FakeMember(FakeModel):

    def __init__(self, admin_state_up=True, pool=None,
                 id='fake-member-id-001',
                 address='2.2.2.2'):
        super(FakeMember, self).__init__()
        self.id = id
        self.address = address
        self.admin_state_up = admin_state_up
        self.pool = pool
        self.protocol_port = 80


class FakePool(FakeModel):

    def __init__(self, protocol, method, persistence, listener=False,
                 members=[], hm=None):
        super(FakePool, self).__init__()
        self.id = 'fake-pool-id-001'
        self.protocol = protocol
        self.lb_algorithm = method
        if persistence is None:
            self.sessionpersistence = None
        else:
            self.sessionpersistence = FakePersistence(persistence)
        if listener:
            self.listener = FakeListener(protocol, 2222, pool=self,
                                         loadbalancer=FakeLoadBalancer())
        else:
            self.listener = None
        for member in members:
            member.pool = self
        self.members = members
        self.healthmonitor = hm
        if hm is not None:
            self.healthmonitor.pool = self


class FakeHM(FakeModel):

    def __init__(self, prot, pool=None):
        super(FakeHM, self).__init__()
        self.id = 'fake-hm-id-001'
        self.name = 'hm1'
        self.type = prot
        self.delay = 6
        self.timeout = 7
        self.max_retries = 8
        self.http_method = 'GET'
        self.url_path = '/'
        self.expected_codes = '200'
        self.pool = pool


class UnitTestBase(test_base.UnitTestBase):

    def __init__(self, *args):
        super(UnitTestBase, self).__init__(*args)
        self.version = 'v2'



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import a10_neutron_lbaas.a10_exceptions as a10_ex

import acos_client.errors as acos_errors
import handler_base
import v1_context as a10

LOG = logging.getLogger(__name__)


class VipHandler(handler_base.HandlerBase):

    def _model_type(self):
        return 'vip'

    def _pool_get(self, context, pool_id):
        return self.openstack_driver.plugin.get_pool(context, pool_id)

    def _pool_name(self, context, pool_id):
        pool = self._pool_get(context, pool_id)
        return self.meta(pool, 'name', pool['id'])

    def _protocols(self, c):
        z = c.client.slb.virtual_server.vport
        return {
            'TCP': z.TCP,
            'UDP': z.UDP,
            'HTTP': z.HTTP,
            'HTTPS': z.TCP,
            'TERMINATED_HTTPS': z.HTTPS,
            'OTHERS': z.OTHERS,
            'RTSP': z.RTSP,
            'FTP': z.FTP,
            'MMS': z.MMS,
            'SIP': z.SIP,
            'FAST_HTTP': z.FAST_HTTP,
            'GENERIC_PROXY': z.GENERIC_PROXY,
            'SSL_PROXY': z.SSL_PROXY,
            'SMTP': z.SMTP,
            'SIP_TCP': z.SIP_TCP,
            'SIPS': z.SIPS,
            'DIAMETER': z.DIAMETER,
            'DNS_UDP': z.DNS_UDP,
            'TFTP': z.TFTP,
            'DNS_TCP': z.DNS_TCP,
            'RADIUS': z.RADIUS,
            'MYSQL': z.MYSQL,
            'MSSQL': z.MSSQL,
            'FIX': z.FIX,
            'SMPP_TCP': z.SMPP_TCP,
            'SPDY': z.SPDY,
            'SPDYS': z.SPDYS,
            'FTP_PROXY': z.FTP_PROXY
        }

    def _meta_name(self, vip):
        return self.meta(vip, 'name', vip['id'])

    def create(self, context, vip):
        with a10.A10WriteStatusContext(self, context, vip) as c:
            status = c.client.slb.UP
            if not vip['admin_state_up']:
                status = c.client.slb.DOWN

            pool_name = self._pool_name(context, vip['pool_id'])

            p = PersistHandler(c, context, vip, self._meta_name(vip))
            p.create()

            templates = self.meta(vip, "template", {})

            if 'client_ssl' in templates:
                args = {'client_ssl_template': templates['client_ssl']}
                try:
                    c.client.slb.template.client_ssl.create(
                        '', '', '',
                        axapi_args=args)
                except acos_errors.Exists:
                    pass

            if 'server_ssl' in templates:
                args = {'server_ssl_template': templates['server_ssl']}
                try:
                    c.client.slb.template.server_ssl.create(
                        '', '', '',
                        axapi_args=args)
                except acos_errors.Exists:
                    pass

            vport_list = [{}]
            try:
                vip_args = {
                    'virtual_server': self.meta(vip, 'virtual_server', {})
                }
                vport_list = vip_args['virtual_server'].pop('vport_list', [{}])
                c.client.slb.virtual_server.create(
                    self._meta_name(vip),
                    vip['address'],
                    status,
                    axapi_args=vip_args)
            except acos_errors.Exists:
                pass

            LOG.debug("VPORT_LIST = %s", vport_list)
            try:
                if vport_list[0]:
                    vport_args = {'vport': vport_list[0]}
                else:
                    vport_args = {'vport': self.meta(vip, 'vport', {})}
                c.client.slb.virtual_server.vport.create(
                    self._meta_name(vip),
                    self._meta_name(vip) + '_VPORT',
                    protocol=self._protocols(c)[vip['protocol']],
                    port=vip['protocol_port'],
                    service_group_name=pool_name,
                    s_pers_name=p.s_persistence(),
                    c_pers_name=p.c_persistence(),
                    status=status,
                    axapi_args=vport_args)
            except acos_errors.Exists:
                pass

            i=1
            for vport in vport_list[1:]:
                i += 1
                try:
                    vport_args = {'vport': vport}
                    c.client.slb.virtual_server.vport.create(
                        self._meta_name(vip),
                        self._meta_name(vip) + '_VPORT' + str(i),
                        protocol=self._protocols(c)[vip['protocol']],
                        port=vip['protocol_port'],
                        service_group_name=pool_name,
                        s_pers_name=p.s_persistence(),
                        c_pers_name=p.c_persistence(),
                        status=status,
                        axapi_args=vport_args)
                except acos_errors.Exists:
                    pass

    def update(self, context, old_vip, vip):
        with a10.A10WriteStatusContext(self, context, vip) as c:
            status = c.client.slb.UP
            if not vip['admin_state_up']:
                status = c.client.slb.DOWN

            pool_name = self._pool_name(context, vip['pool_id'])

            p = PersistHandler(c, context, vip, self._meta_name(vip))
            p.create()

            templates = self.meta(vip, "template", {})

            if 'client_ssl' in templates:
                args = {'client_ssl_template': templates['client_ssl']}
                c.client.slb.template.client_ssl.update(
                    '', '', '',
                    axapi_args=args)

            if 'server_ssl' in templates:
                args = {'server_ssl_template': templates['server_ssl']}
                c.client.slb.template.server_ssl.update(
                    '', '', '',
                    axapi_args=args)

            vport_args = {'vport': self.meta(vip, 'vport', {})}
            c.client.slb.virtual_server.vport.update(
                self._meta_name(vip),
                self._meta_name(vip) + '_VPORT',
                protocol=self._protocols(c)[vip['protocol']],
                port=vip['protocol_port'],
                service_group_name=pool_name,
                s_pers_name=p.s_persistence(),
                c_pers_name=p.c_persistence(),
                status=status,
                axapi_args=vport_args)

    def _delete(self, c, context, vip):
        c.client.slb.virtual_server.delete(self._meta_name(vip))
        PersistHandler(c, context, vip, self._meta_name(vip)).delete()

    def delete(self, context, vip):
        with a10.A10DeleteContext(self, context, vip) as c:
            self._delete(c, context, vip)


class PersistHandler(object):

    def __init__(self, c, context, vip, vip_name):
        self.c = c
        self.context = context
        self.vip = vip
        self.c_pers = None
        self.s_pers = None
        self.name = vip_name

        if vip.get('session_persistence', None) is not None:
            self.sp = vip['session_persistence']
            if self.sp['type'] == 'HTTP_COOKIE':
                self.c_pers = self.name
            elif self.sp['type'] == 'SOURCE_IP':
                self.s_pers = self.name
            else:
                raise a10_ex.UnsupportedFeature()
        else:
            self.sp = None

    def c_persistence(self):
        return self.c_pers

    def s_persistence(self):
        return self.s_pers

    def create(self):
        if self.sp is None:
            return

        methods = {
            'HTTP_COOKIE':
                self.c.client.slb.template.cookie_persistence.create,
            'SOURCE_IP':
                self.c.client.slb.template.src_ip_persistence.create,
        }
        if self.sp['type'] in methods:
            try:
                methods[self.sp['type']](self.name)
            except acos_errors.Exists:
                pass

    def delete(self):
        if self.sp is None:
            return

        methods = {
            'HTTP_COOKIE':
                self.c.client.slb.template.cookie_persistence.delete,
            'SOURCE_IP':
                self.c.client.slb.template.src_ip_persistence.delete,
        }
        if self.sp['type'] in methods:
            try:
                methods[self.sp['type']](self.name)
            except Exception:
                pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.a10_exceptions as a10_ex

import handler_base
import v2_context as a10


class ListenerHandler(handler_base.HandlerBase):

    def _protocols(self, c):
        return {
            'TCP': c.client.slb.virtual_server.vport.TCP,
            'UDP': c.client.slb.virtual_server.vport.UDP,
            'HTTP': c.client.slb.virtual_server.vport.HTTP,
            'HTTPS': c.client.slb.virtual_server.vport.TCP
        }

    def _persistence_get(self, c, context, listener):
        if (not listener.default_pool or
           not listener.default_pool.sessionpersistence):
            return [None, None]

        sp = listener.default_pool.sessionpersistence
        name = listener.default_pool.id
        c_pers = None
        s_pers = None

        if sp.type == 'HTTP_COOKIE':
            c_pers = name
        elif sp.type == 'SOURCE_IP':
            s_pers = name
        else:
            raise a10_ex.UnsupportedFeature()

        return [c_pers, s_pers]

    def _set(self, c, set_method, context, listener):
        status = c.client.slb.UP
        if not listener.admin_state_up:
            status = c.client.slb.DOWN

        pers = self._persistence_get(c, context, listener)

        set_method(listener.loadbalancer.id, listener.id,
                   protocol=self._protocols(c)[listener.protocol],
                   port=listener.port,
                   service_group_name=listener.default_pool.id,
                   s_pers_name=pers[1],
                   c_pers_name=pers[0],
                   status=status)

    def _create(self, c, context, listener):
        self._set(c, c.client.slb.virtual_server.vport.create, context,
                  listener)

    def create(self, context, listener):
        if not listener.loadbalancer or not listener.default_pool:
            return

        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._create(c, context, listener)

    def _update(self, c, context, listener):
        self._set(c, c.client.slb.virtual_server.vport.update, context,
                  listener)

    def update(self, context, old_listener, listener):
        if not listener.loadbalancer or not listener.default_pool:
            return

        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._update(c, context, listener)

    def delete(self, context, listener):
        if not listener.loadbalancer:
            return

        with a10.A10DeleteContext(self, context, listener) as c:
            c.client.slb.virtual_server.vport.delete(
                listener.loadbalancer.id,
                listener.id,
                protocol=self._protocols(c)[listener.protocol],
                port=listener.port)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import a10_config
import acos_client
import v1.handler_hm
import v1.handler_member
import v1.handler_pool
import v1.handler_vip
import v2.handler_hm
import v2.handler_lb
import v2.handler_listener
import v2.handler_member
import v2.handler_pool
import version

LOG = logging.getLogger(__name__)


class A10OpenstackLBBase(object):

    def __init__(self, openstack_driver):
        self.openstack_driver = openstack_driver
        self.config = a10_config.A10Config()
        self.appliance_hash = acos_client.Hash(self.config.devices.keys())
        if self.config.verify_appliances:
            self._verify_appliances()

        LOG.info("A10-neutron-lbaas: initializing, version=%s, acos_client=%s",
                 version.VERSION, acos_client.VERSION)

    def _select_a10_device(self, tenant_id):
        s = self.appliance_hash.get_server(tenant_id)
        return self.config.devices[s]

    def _get_a10_client(self, device_info):
        d = device_info
        return acos_client.Client(d['host'],
                                  d.get('api_version', acos_client.AXAPI_21),
                                  d['username'], d['password'],
                                  port=d['port'], protocol=d['protocol'])

    def _verify_appliances(self):
        LOG.info("A10Driver: verifying appliances")

        if len(self.config.devices) == 0:
            LOG.error("A10Driver: no configured appliances")

        for k, v in self.config.devices.items():
            try:
                LOG.info("A10Driver: appliance(%s) = %s", k,
                         self._get_a10_client(v).system.information())
            except Exception:
                LOG.error("A10Driver: unable to connect to configured"
                          "appliance, name=%s", k)


class A10OpenstackLBV2(A10OpenstackLBBase):

    @property
    def lb(self):
        return v2.handler_lb.LoadBalancerHandler(
            self,
            self.openstack_driver.load_balancer)

    @property
    def listener(self):
        return v2.handler_listener.ListenerHandler(
            self,
            self.openstack_driver.listener)

    @property
    def pool(self):
        return v2.handler_pool.PoolHandler(self, self.openstack_driver.pool)

    @property
    def member(self):
        return v2.handler_member.MemberHandler(
            self,
            self.openstack_driver.member)

    @property
    def hm(self):
        return v2.handler_hm.HealthMonitorHandler(
            self,
            self.openstack_driver.health_monitor)


class A10OpenstackLBV1(A10OpenstackLBBase):

    @property
    def pool(self):
        return v1.handler_pool.PoolHandler(self)

    @property
    def vip(self):
        return v1.handler_vip.VipHandler(self)

    @property
    def member(self):
        return v1.handler_member.MemberHandler(self)

    @property
    def hm(self):
        return v1.handler_hm.HealthMonitorHandler(self)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
import handler_base
import v1_context as a10


class MemberHandler(handler_base.HandlerBase):

    def _model_type(self):
        return 'member'

    def _get_ip(self, context, member, use_float=False):
        return self.openstack_driver._member_get_ip(context, member, use_float)

    def _pool_get(self, context, pool_id):
        return self.openstack_driver.plugin.get_pool(context, pool_id)

    def _pool_name(self, context, pool_id):
        pool = self._pool_get(context, pool_id)
        return self.meta(pool, 'name', pool['id'])

    def _get_name(self, member, ip_address):
        tenant_label = member['tenant_id'][:5]
        addr_label = str(ip_address).replace(".", "_", 4)
        server_name = "_%s_%s_neutron" % (tenant_label, addr_label)
        return server_name

    def _meta_name(self, member, ip_address):
        return self.meta(member, 'name', self._get_name(member, ip_address))

    def _count(self, context, member):
        return self.openstack_driver._member_count(context, member)

    def _create(self, c, context, member):
        server_ip = self._get_ip(context, member,
                                 c.device_cfg['use_float'])
        server_name = self._meta_name(member, server_ip)

        status = c.client.slb.UP
        if not member['admin_state_up']:
            status = c.client.slb.DOWN

        try:
            server_args = {'server': self.meta(member, 'server', {})}
            c.client.slb.server.create(server_name, server_ip,
                                       axapi_args=server_args)
        except (acos_errors.Exists, acos_errors.AddressSpecifiedIsInUse):
            pass

        try:
            member_args = {'member': self.meta(member, 'member', {})}
            c.client.slb.service_group.member.create(
                self._pool_name(context, member['pool_id']),
                server_name,
                member['protocol_port'],
                status=status,
                axapi_args=member_args)
        except acos_errors.Exists:
            pass

    def create(self, context, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            self._create(c, context, member)

    def update(self, context, old_member, member):
        with a10.A10WriteStatusContext(self, context, member) as c:
            server_ip = self._get_ip(context, member,
                                     c.device_cfg['use_float'])
            server_name = self._meta_name(member, server_ip)

            status = c.client.slb.UP
            if not member['admin_state_up']:
                status = c.client.slb.DOWN

            try:
                member_args = {'member': self.meta(member, 'member', {})}
                c.client.slb.service_group.member.update(
                    self._pool_name(context, member['pool_id']),
                    server_name,
                    member['protocol_port'],
                    status,
                    axapi_args=member_args)
            except acos_errors.NotFound:
                # Adding db relation after the fact
                self._create(c, context, member)

    def _delete(self, c, context, member):
        server_ip = self._get_ip(context, member, c.device_cfg['use_float'])
        server_name = self._meta_name(member, server_ip)

        try:
            if self._count(context, member) > 1:
                c.client.slb.service_group.member.delete(
                    self._pool_name(context, member['pool_id']),
                    server_name,
                    member['protocol_port'])
            else:
                c.client.slb.server.delete(server_name)
        except acos_errors.NotFound:
            pass

    def delete(self, context, member):
        with a10.A10DeleteContext(self, context, member) as c:
            self._delete(c, context, member)
